/*
 * sata_hotplug_spd_fix.c
 *
 * Uses a kretprobe on sata_down_spd_limit to prevent libata from limiting SATA
 * speed due to hotplugging.
 *
 * Based on samples/kprobes/kretprobe_example.c from the Linux source tree.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kprobes.h>
#include <linux/ktime.h>
#include <linux/limits.h>
#include <linux/sched.h>
#include <linux/libata.h>

static char func_name[NAME_MAX] = "sata_down_spd_limit";

/* per-instance private data */
struct my_data {
    struct ata_link* link;
    unsigned int orig_spd_limit;
};

static int entry_handler(struct kretprobe_instance *ri, struct pt_regs *regs)
{
	struct my_data *data;
        struct ata_link* link;
	data = (struct my_data *)ri->data;

#ifdef __i386__
        /* gcc -mregparm puts arg 1 in eax */
        link = (struct ata_link*)regs->ax;
#elif defined(__x86_64__)
        link = (struct ata_link*)regs->di;
#else
#error "unknown arch"
#endif


        if (!link) {
            printk(KERN_INFO "sata_down_spd_limit called with null ata_link\n");
            return -EIO;
        }

        data->link = link;
        data->orig_spd_limit = link->sata_spd_limit;

	return 0;
}

static int ret_handler(struct kretprobe_instance *ri, struct pt_regs *regs)
{
    struct my_data *data;
    data = (struct my_data *)ri->data;
    printk(KERN_INFO "sata_down_spd_limit set sata_spd_limit to %x, resetting to %x\n", data->link->sata_spd_limit, data->orig_spd_limit);
    data->link->sata_spd_limit = data->orig_spd_limit;
    return 0;
}

static struct kretprobe my_kretprobe = {
	.handler		= ret_handler,
	.entry_handler		= entry_handler,
	.data_size		= sizeof(struct my_data),
	/* Probe up to 20 instances concurrently. */
	.maxactive		= 20,
};

static int __init sata_hotplug_spd_fix_init(void)
{
	int ret;

	my_kretprobe.kp.symbol_name = func_name;
	ret = register_kretprobe(&my_kretprobe);
	if (ret < 0) {
		printk(KERN_INFO "sata_hotplug_spd_fix: register_kretprobe failed, returned %d\n",
				ret);
		return -1;
	}
	printk(KERN_INFO "sata_hotplug_spd_fix: Planted return probe at %s: %p\n",
			my_kretprobe.kp.symbol_name, my_kretprobe.kp.addr);
	return 0;
}

static void __exit sata_hotplug_spd_fix_exit(void)
{
	unregister_kretprobe(&my_kretprobe);
	printk(KERN_INFO "sata_hotplug_spd_fix: kretprobe at %p unregistered\n",
			my_kretprobe.kp.addr);
}

module_init(sata_hotplug_spd_fix_init)
module_exit(sata_hotplug_spd_fix_exit)
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Katie Stafford <katie@ktpanda.org>");
MODULE_DESCRIPTION("Prevents libata from limiting SATA speed due to hotplugging");
MODULE_VERSION("1.0");
